!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module ps_generation
  use global
  use oct_parser
  use io
  use messages
  use units
  use output
  use mesh
  use quantum_numbers
  use potentials
  use wave_equations_integrator
  use eigensolver_m
  use hamann
  use states
  use hartree
  use xc
  use ps_io
  implicit none


                    !---Derived Data Types---!

  type ps_generator_type
    integer :: nspin
    integer :: scheme
    real(R8) :: tol
    character(3) :: unbound_trick
    integer :: n_states, n_sc
    type(qn_type), pointer :: qn(:,:)
    real(R8), pointer :: rc(:)
    !The quantum numbers and core radii for each state should be ordered in the
    !following way:
    ! * qn(:, 1:n_sc) - the semi-core states.
    ! * qn(:, n_sc + 1:n_states - n_sc) - the valence states which do NOT 
    !   have a corresponding semi-core state.
    ! * qn(n_states - n_sc + 1:n_states) - the valence states which have a 
    !   corresponding semi-core state ordered in the same way as the semi-core
    !   states.
  end type ps_generator_type


                    !---Global Variables---!



                    !---Public/Private Statements---!

  private
  public :: ps_generator_type, &
            ps_generator_null, &
            ps_generator_init, &
            ps_generate, &
            ps_generate_kb, &
            ps_generator_end

contains

  subroutine ps_generator_null(ps_gen)
    !-----------------------------------------------------------------------!
    ! Nullifies all the components of ps_generator.                         !
    !-----------------------------------------------------------------------!
    type(ps_generator_type), intent(out) :: ps_gen

    call push_sub("ps_generator_null")

    nullify(ps_gen%qn)
    nullify(ps_gen%rc)

    call pop_sub()
  end subroutine ps_generator_null

  subroutine ps_generator_init(nspin, wave_eq, ps_gen, n_states, n_sc)
    !-----------------------------------------------------------------------!
    ! Initialize the pseudo-atom: read the input options and get the        !
    ! necessary information from the all-electron atom.                     !
    !-----------------------------------------------------------------------!
    integer,                 intent(in)    :: nspin, wave_eq
    type(ps_generator_type), intent(inout) :: ps_gen
    integer,                 intent(out)   :: n_states, n_sc

    integer :: n_v, i, j, j1, j2, l
    real(R8) :: sc_rc(7), v_rc(7)
    type(qn_type) :: sc_qn(7), v_qn(7)

    call push_sub("ps_atom_init")

    ps_gen%nspin = nspin

    !What scheme are we using to generate the PP?
    call oct_parse_int('PPScheme', HAM, ps_gen%scheme)
    select case (ps_gen%scheme)
    case (HAM,TM,MRPP,MTM)
    case (RTM)
      if (wave_eq /= DIRAC) then
        message(1) = "All electron relativistic calculation is needed in order to use"
        message(2) = "the relativistic extension to the Troullier-Martins scheme ."
        call write_fatal(2)
      end if
    case (RMRPP)
      if (wave_eq /= DIRAC) then
        message(1) = "All electron relativistic calculation is needed in order."
        message(2) = "to use the relativistic extension to the MRPP scheme"
        call write_fatal(2)
      end if
    case default
      message(1) = "Illegal pseusopotential generation scheme."
      call write_fatal(1)
    end select

    !Get PP generation tolerance
    call oct_parse_double('PPCalculationTolerance', 1e-6_r8, ps_gen%tol)
    if (ps_gen%tol <= M_ZERO) then
      message(1) = "PPCalculationTolerance must be positive."
      call write_fatal(1)
    end if

    !Check how to deal with the unbound states
    call oct_parse_string('PPTrick', 'non', ps_gen%unbound_trick)

    !Start by the semi-core components
    if (oct_parse_isdef('PPSemiCoreComponents') /= 0) then
      call read_input_data("PPSemiCoreComponents", ps_gen%n_sc, sc_qn, sc_rc)
    else
      ps_gen%n_sc = 0
    end if

    !Next the valence
    call read_input_data("PPComponents", n_v, v_qn, v_rc)

    !If we have semi-core states, check a couple of things
    if (ps_gen%n_sc /= 0) then
      do l = 0, 3
        if (any(sc_qn%l == l) .and. .not. any(v_qn%l == l)) then
          message(1) = "Some of the angular channels present in the PPSemiCoreComponents"
          message(2) = "block are not present in the PPComponents block"
          call write_fatal(2)
        end if
      end do
    end if

    !Get how many pseudo-states we will have and allocate the memory
    ps_gen%n_states = ps_gen%n_sc + n_v
    allocate(ps_gen%qn(nspin, ps_gen%n_states))
    allocate(ps_gen%rc(ps_gen%n_states))

    !Initialize semi-core states
    do i = 1, ps_gen%n_sc
      ps_gen%rc(i) = sc_rc(i)
      ps_gen%qn(:, i) = sc_qn(i)
      if (nspin == 2) then
        ps_gen%qn(1, i)%s = - M_HALF
        ps_gen%qn(2, i)%s =   M_HALF
      end if
    end do

    !Initialize valence states
    j1 = ps_gen%n_sc + 1
    j2 = ps_gen%n_states - ps_gen%n_sc + 1
    do i = 1, n_v

      if (ps_gen%n_sc /= 0) then
        if (any(sc_qn%l == v_qn(i)%l .and. sc_qn%j == v_qn(i)%j)) then
          j = j2
          j2 = j2 + 1
        else
          j = j1
          j1 = j1 + 1
        end if
      else
        j = i
      end if

      ps_gen%rc(j) = v_rc(i)
      ps_gen%qn(:, j) = v_qn(i)
      if (nspin == 2) then
        ps_gen%qn(1, j)%s = - M_HALF
        ps_gen%qn(2, j)%s =   M_HALF
      end if
    end do

    n_states = ps_gen%n_states
    n_sc = ps_gen%n_sc

    call pop_sub()
  contains
    
    subroutine read_input_data(blk_name, n_block, qn, rc)
      character(len=*), intent(in)  :: blk_name
      integer,          intent(out) :: n_block
      type(qn_type),    intent(out) :: qn(7)
      real(R8),         intent(out) :: rc(7)

      integer :: i, k, ierr, n_cols, n_lines, n, l
      integer(POINTER_SIZE) :: blk
      real(R8) :: u, j

      qn = QN_NULL

      !Open the block and how many lines the block has
      ierr = oct_parse_block(trim(blk_name), blk)
      n_lines = oct_parse_block_n(blk)

      !Read the input data
      u = units_in%length%factor
      k = 0
      do i = 1, n_lines
        k = k + 1
        n_cols = oct_parse_block_cols(blk, i-1)

        !Check for errors in the format
        if (n_cols < 3 .or. n_cols > 4 .or. &
            (wave_eq /= DIRAC .and. n_cols == 4)) then
          write(message(1),'("There is something wrong at line ",I2," of ",A," block.")') i, trim(blk_name)
          call write_fatal(1)
        end if

        !Get main quantum number
        call oct_parse_block_int(blk, i-1, 0, n)

        !Get angular momentum quantum number
        call oct_parse_block_int(blk, i-1, 1, l)

        !Get cutoff radius (it is always the last column)
        call oct_parse_block_double(blk, i-1, n_cols-1, rc(k))
        rc(k) = rc(k)*u

        if (wave_eq == DIRAC) then
          if (n_cols == 4) then
            !The second line is the total angular momentum quantum number
            call oct_parse_block_double(blk, i-1, 2, j)
            qn(k) = qn_init(n, l, M_ZERO, j)

          else 
            if (l == 0) then
              j = l + M_HALF
              qn(k) = qn_init(n, l, M_ZERO, j)
            else
              j = l - M_HALF
              qn(k) = qn_init(n, l, M_ZERO, j)
              j = l + M_HALF
              qn(k + 1) = qn_init(n, l, M_ZERO, j)
              rc(k + 1) = rc(k)
              k = k + 1
            end if

          end if
        else
          j = M_ZERO
          qn(k) = qn_init(n, l, M_ZERO, j)
        end if

      end do
      n_block = k

      !Check for errors in the input data
      do i = 1, n        
        if (rc(i) < M_ZERO) then
          write(message(1),'("There is something wrong at line ",I2," of ",A," block.")') i, trim(blk_name)
          message(2) = "The cutoff radius can not take negative values."
          call write_fatal(2)
        end if
      end do

      do i = 1, n_block
        do k = i + 1, n_block
          if (qn(i) == qn(k)) then
            write(message(1),'("Some lines in the ",A," ")') trim(blk_name)
            message(2) = "block have the same quantum numbers."
            call write_fatal(2)
          end if
        end do
      end do

      call oct_parse_block_end(blk)

    end subroutine read_input_data

  end subroutine ps_generator_init

  subroutine ps_generate(ps_gen, m, wave_eq, nspin, n_ae_states, ae_states, &
                         ae_potential, xc_model, integrator_sp, integrator_dp, &
                         eigensolver, ps_states, ps_potential)
    !-----------------------------------------------------------------------!
    ! Generate  the pseudo-potentials.                                      !
    !-----------------------------------------------------------------------!
    type(ps_generator_type), intent(inout) :: ps_gen
    type(mesh_type),         intent(in)    :: m
    integer,                 intent(in)    :: wave_eq, nspin, n_ae_states
    type(state_type),        intent(in)    :: ae_states(nspin, n_ae_states)
    type(potential_type),    intent(in)    :: ae_potential
    type(xc_type),           intent(inout) :: xc_model
    type(integrator_type),   intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type),  intent(in)    :: eigensolver
    type(state_type),        intent(inout) :: ps_states(nspin, ps_gen%n_states)
    type(potential_type),    intent(inout) :: ps_potential

    integer :: p, i, n, nlcc, v_dim
    real(R8) :: max_ev, rc, z1, z2, ev, norm, slope, rmatch
    character(len=5) :: label
    real(R8), allocatable :: ps_v(:,:,:), vv(:,:)
    real(R8), allocatable :: core_density(:,:), ps_density(:,:)
    real(R8), allocatable :: vh(:,:), vxc(:,:), vxctau(:,:)
    type(qn_type), allocatable :: qn(:)

    call push_sub("ps_generation")

    !Print some information
    message(1) = ""
    message(2) = "Pseudo atom information:"
    call write_info(2)
    call write_info(2,unit=out_pp)


          !---Get the all-electron states and check the core radii---!

    do p = 1, nspin
      do i = 1, ps_gen%n_states
        do n = 1, n_ae_states
          if (state_qn(ae_states(p, n)) == ps_gen%qn(p, i)) then
            ps_states(p, i) = ae_states(p, n)
            exit
          end if
        end do
      end do
    end do

    !Check core radius
    do i = 1, ps_gen%n_states
      if (ps_gen%rc(i) == M_ZERO) ps_gen%rc(i) = &
                                 state_default_rc(ps_states(1, i), ps_gen%scheme)
      if (ps_gen%rc(i) == M_ZERO) ps_gen%rc(i) = &
                               maxval(state_default_rc(ps_states, ps_gen%scheme))
    end do

    !Do something about the unbound states
    if (ps_gen%unbound_trick == "fhi") then
      max_ev = maxval(state_eigenvalue(ae_states), &
                      state_eigenvalue(ae_states) /= M_ZERO)
      do p = 1, nspin
        do i = 1, ps_gen%n_states
          if (state_eigenvalue(ps_states(p, i)) == M_ZERO) then
            ps_states(p, i)%ev = max_ev
          end if
        end do
      end do
    end if

    !Get core density
    allocate(core_density(m%np, nspin))
    core_density = state_density(nspin, n_ae_states, ae_states, m) - &
                   state_density(nspin, ps_gen%n_states, ps_states, m)


    !Write information
    call ps_generation_output_input(ps_gen, ps_states, verbose_limit=20)
    call ps_generation_output_input(ps_gen, ps_states, out_pp)
    message(1) = ""
    message(2) = "Pseudopotential Generation:"
    call write_info(2)
    call write_info(2,unit=out_pp)


          !---Generate pseudopotentials---!

    v_dim = ps_gen%n_states - ps_gen%n_sc
    allocate(ps_v(m%np, nspin, v_dim))
    ps_v = M_ZERO
    do p = 1, nspin
      do i = 1, ps_gen%n_sc
        call state_psp_generation(m, ps_gen%scheme, wave_eq, ps_gen%tol, &
             ae_potential, integrator_sp, integrator_dp, ps_v(:, p, i), &
             ps_states(p, i), ps_gen%rc(i), &
             ps_states(p, ps_gen%n_states - ps_gen%n_sc + i), &
             ps_gen%rc(ps_gen%n_states - ps_gen%n_sc + i))
      end do
      do i = ps_gen%n_sc + 1, ps_gen%n_states - ps_gen%n_sc
        call state_psp_generation(m, ps_gen%scheme, wave_eq, ps_gen%tol, &
             ae_potential, integrator_sp, integrator_dp, ps_v(:, p, i), &
             ps_states(p, i), ps_gen%rc(i))
      end do
    end do


        !---Unscreening and spin-average of the pseudopotentials---!


    !Non-linear core corrections
    call oct_parse_int('CoreCorrection', M_CC_NONE, nlcc)
    if (nlcc /= M_CC_NONE) then
      !Get rc
      allocate(ps_density(m%np, nspin))
      ps_density = state_density(nspin, ps_gen%n_states, ps_states, m)
      do i = m%np, 2, -1 
        if (m%r(i)**2*sum(core_density(i, 1:nspin)) > m%r(i)**2*sum(ps_density(i, 1:nspin))) then
          rc = m%r(i)
          exit
        end if
      end do
      deallocate(ps_density)

      if (nlcc == M_CC_FHI) then
        call oct_parse_double('CoreCorrectionCutoff', m%r(i), rc)
      endif
    
      !Initialize the nlcc part of xc_model
      call xc_nlcc_init(xc_model, m, rc, nlcc, core_density)
    end if
    deallocate(core_density)

    !Get screening
    allocate(vh(m%np, nspin), vxc(m%np, nspin), vxctau(m%np, nspin))
    call hartree_potential(m, nspin, state_density(nspin, ps_gen%n_states, ps_states, m), vh=vh)
    call xc_potential(xc_model, m, nspin, ps_gen%n_states, ps_states, vxc=vxc, vxctau=vxctau)

    !Unscreen
    do p = 1, nspin
      do i = 1, v_dim
        ps_v(:, p, i) = ps_v(:, p, i) - vh(:, p) - vxc(:, p)
      end do
    end do

   !Initialize ps_potential structure
    allocate(vv(m%np, v_dim), qn(v_dim))
    do i = 1, v_dim
      qn(i) = state_qn(ps_states(1, i))

      if (nspin == 1) then
        vv(:, i) = ps_v(:, 1, i)
      else
        !We will assume that core polarization is zero. This means we need to
        !convert the polarized potentials back to non-polarized potentials
        z1 = state_charge(ps_states(1, i))
        z2 = state_charge(ps_states(2, i))
        if (ps_gen%n_sc /= 0) then
          z1 = z1 + state_charge(ps_states(1, v_dim + i))
          z2 = z2 + state_charge(ps_states(2, v_dim + i))
        end if

        if (z1 + z2 == M_ZERO) then
          z1 = 1 ; z2 = 1
        end if
        vv(:, i) = (ps_v(:, 1, i)*z1 + ps_v(:, 2, i)*z2)/(z1 + z2)

      end if

    end do 
    deallocate(ps_v)

    call potential_init(ps_potential, m, nspin, vh + vxc, vxctau, v_dim, qn, vv)
    deallocate(vv, qn, vh, vxc, vxctau)


          !---Self-consistency test---!

    message(1) = ""
    message(2) = "Pseudopotentials Self-Consistency:"
    write(message(3),'(2X,"State  Eigenvalue [",A,"]    Norm Test   Slope Test")') units_out%energy%abbrev
    call write_info(3)
    write(out_pp,*)
    call write_info(3,unit=out_pp)
    do p = 1, nspin
      do i = 1, ps_gen%n_states
        if (ps_gen%scheme == HAM) then
          rmatch = m%r(hamann_match_radius(m, ps_gen%rc(i)))
        else
          rmatch = ps_gen%rc(i)
        end if

        call state_test_consistency(m, wave_eq, eigensolver, integrator_sp, &
             integrator_dp, ae_potential, ps_potential, ps_states(p, i), &
             rmatch, ev, norm, slope)

        label = state_label(ps_states(p, i), spin=.true.)
        select case (len_trim(label))
        case (2)
          write(message(1),'(4X,A,4X,F12.5,5X,F10.7,2X,F10.7)') trim(label), &
                                    ev/units_out%energy%factor, norm, slope
        case (5)
          write(message(1),'(2X,A,3X,F12.5,5X,F10.7,2X,F10.7)') trim(label), &
                                    ev/units_out%energy%factor, norm, slope
        end select
        call write_info(1)
        call write_info(1,unit=out_pp)
      end do
    end do

    call pop_sub()
  end subroutine ps_generate

  subroutine ps_generate_kb(ps_gen, m, wave_eq, nspin, n_states, n_sc, &
                          ps_states, z, xc_model, ps_potential, integrator_sp, &
                          integrator_dp, eigensolver, kb_potential)
    !-----------------------------------------------------------------------!
    ! Generate the KB form of the pseudo-potentials.                        !
    !-----------------------------------------------------------------------!
    type(ps_generator_type), intent(inout) :: ps_gen
    type(mesh_type),         intent(in)    :: m
    integer,                 intent(in)    :: wave_eq, nspin, n_states, n_sc
    type(state_type),        intent(in)    :: ps_states(nspin, n_states)
    real(R8),                intent(in)    :: z
    type(xc_type),           intent(inout) :: xc_model
    type(potential_type),    intent(in)    :: ps_potential
    type(integrator_type),   intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type),  intent(in)    :: eigensolver
    type(potential_type),    intent(inout) :: kb_potential

    integer :: l_local, i, p, k, n_kb
    real(R8) :: rcmax, v0, v1, v2, v3, r
    real(R8), allocatable :: vh(:,:), vxc(:,:), vxctau(:,:), vv(:), e(:), proj_f(:,:)
    type(qn_type) :: qn_loc
    type(qn_type), allocatable :: qn(:)
    type(potential_type) :: v_local

    call push_sub("ps_generate_kb")

    call potential_null(v_local)

    !Write information
    message(1) = ""
    message(2) = "Kleinman & Bylander Atom"
    call write_info(2)
    call write_info(2, unit=out_kb)

    !Read input
    call oct_parse_int("Llocal", -1, l_local)
    if (l_local < -1) then
      message(1) = "Invalid value for Llocal."
      call write_fatal(1)
    end if
    if (l_local == -1) then
      write(message(1),'(2X,"Local potential is a Vanderbilt function")')
    else
      write(message(1),'(2X,"l-component used as local: l = ",I1)') l_local
    end if
    call write_info(1, 20)
    call write_info(1, unit=out_kb)

    ! Get local potential
    if (l_local == -1) then
      ! If llocal=-1, use a Vanderbilt function for the local component
      rcmax = maxval(ps_gen%rc)
      v0 = -z*exp(M_ELEVEN/M_TWELVE)/rcmax
      v1 = -M_THREE/(M_TWO*rcmax**2)
      v2 = M_THREE/(M_FOUR*rcmax**4)
      v3 = -M_ONE/(M_SIX*rcmax**6)

      allocate(vv(m%np))
      do i = 1, m%np
        if (m%r(i) < rcmax) then
          vv(i) = v0*exp(v1*m%r(i)**2 + v2*m%r(i)**4 + v3*m%r(i)**6)
        else
          vv(i) = -z/m%r(i)
        end if
      end do

      !Write some information to the screen
      write(message(1),'(4X,3X,"z",5X,"rcmax",6X,"v0",10X,"v1",10X,"v2",10X,"v3")')
      write(message(2),'(4X,F6.2,2X,F6.2,2X,4(F10.6,2X))') z, rcmax, v0, v1, v2, v3
      call write_info(2)
      call write_info(2, unit=out_kb)

    else
      allocate(vv(m%np))
      qn_loc = QN_NULL
      qn_loc%l = l_local
      if (wave_eq == DIRAC .and. l_local /= 0) then
        qn_loc%j = l_local - M_HALF
        do k = 1, m%np
          vv(k) = v(ps_potential, m%r(k), qn_loc, unscreened=.true.)
        end do
        qn_loc%j = l_local + M_HALF
        do k = 1, m%np
          vv(k) = M_ONE/(M_TWO*l_local + M_ONE)*(l_local*vv(k) + &
            (l_local + M_ONE)*v(ps_potential, m%r(k), qn_loc, unscreened=.true.))
        end do
        
      else
        do k = 1, m%np
          vv(k) = v(ps_potential, m%r(k), qn_loc, unscreened=.true.)
        end do
      end if
    end if

    !Get screening and initialize local potential
    allocate(vh(m%np, nspin), vxc(m%np, nspin), vxctau(m%np, nspin))
    call hartree_potential(m, nspin, state_density(nspin, n_states, ps_states, m), vh=vh)
    call xc_potential(xc_model, m, nspin, n_states, ps_states, vxc=vxc, vxctau=vxctau)
    call potential_init(v_local, m, nspin, vh + vxc, vxctau, vv)
    deallocate(vv, vh, vxc, vxctau)
    write(message(1),'(2X,"Non-local components:")')
    write(message(2),'(4X,"State   KB Energy [",A,"]   KB Cosine")') units_out%energy%abbrev
    call write_info(2)
    call write_info(2, unit=out_kb)

    !Get the projectors and the KB energy
    n_kb = n_states - n_sc
    allocate(qn(n_kb), e(n_kb), proj_f(m%np, n_kb))
    do i = 1, n_kb
      call state_kb_projector(m, nspin, v_local, ps_potential, &
                     ps_states(1:nspin, i), qn(i), e(i), proj_f(:, i))
    end do

    !Initialize the KB potential
    call potential_init(kb_potential, m, l_local, v_local, n_kb, qn, e, proj_f)

    !Test for ghost states
    message(1) = ""
    write(message(2),'(2X,"Ghost state analysis:")')
    call write_info(2)
    call write_info(2, unit=out_kb)
    do p = 1, nspin
      do i = 1, n_states
        call state_test_ghost(m, wave_eq, integrator_sp, integrator_dp, &
                              eigensolver, kb_potential, ps_states(p, i))
      end do
    end do

    !Print localization radii
    message(1) = ""
    write(message(2),'(2X,"Localization radii [",A,"]:")') trim(units_out%length%abbrev)

    r = M_ZERO
    do i = m%np, 1, -1
      if (abs(v(v_local, m%r(i), qn_type(0, 0, M_ZERO, M_ZERO, 0), unscreened=.true.) &
           + sum(state_charge(ps_states))/m%r(i)) > 0.001) then
        r = m%r(i)
        exit
      end if
    end do
    write(message(3),'(4X,"Local:",F6.2)') r/units_out%length%factor

    k = 0
    do p = 1, n_kb
      if (qn(p)%l == l_local) cycle
      k = k + 1
      r = M_ZERO
      do i = m%np, 1, -1
        if (abs(proj_f(i, p)) > 0.001) then
          r = m%r(i)
          exit
        end if
      end do
      write(message(3+k),'(4X,"l = ",I1,":",F6.2)') qn(p)%l, r/units_out%length%factor
    end do
    call write_info(3+k)
    call write_info(3+k, unit=out_kb)

    deallocate(qn, e, proj_f)

    call pop_sub()
  end subroutine ps_generate_kb

  subroutine ps_generator_end(ps_gen)
    !-----------------------------------------------------------------------!
    ! Frees all the memory associated to ps_gen.                            !
    !-----------------------------------------------------------------------!
    type(ps_generator_type), intent(inout) :: ps_gen

    call push_sub("ps_generator_end")

    if (associated(ps_gen%rc)) deallocate(ps_gen%rc)
    if (associated(ps_gen%qn)) deallocate(ps_gen%qn)

    call pop_sub()
  end subroutine ps_generator_end

  subroutine ps_generation_output_input(ps_gen, ps_states, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Writes the pseudo-atom input options and in a nice readable format.   !
    ! If a unit number is provided, it writes it to a file; otherwise it    !
    ! writes it to the standard output.                                     !
    !-----------------------------------------------------------------------!
    type(ps_generator_type), intent(in) :: ps_gen
    type(state_type),        intent(in) :: ps_states(ps_gen%nspin, ps_gen%n_states)
    integer,                 intent(in), optional :: unit, verbose_limit

    integer :: n, i, p
    character(len=5) :: label
    character(len=40) :: scheme

    call push_sub("ps_atom_output_input")

    select case (ps_gen%scheme)
    case (HAM)
      scheme = "Hamann"
    case (TM)
      scheme = "Troullier-Martins"
    case (RTM)
      scheme = "Troullier-Martins Relativistic extension"
    case (MRPP)
      scheme = "Multireference Pseudopotentials (MRPP)"
    case (RMRPP)
      scheme = "MRPP Relativistic extension"
    case (MTM)
      scheme = "Modified Troullier-Martins"
    end select
    write(message(1),'(2X,"PP generation scheme : ",A)') trim(scheme)
    write(message(2),'(2X,"Wavefunction info:")')
    write(message(3),'(4X,"State   Occupation    Node radius   Peak radius   Default core radius")')
    n = 3
    do i = 1, ps_gen%n_states
      do p = 1, ps_gen%nspin
        n = n + 1
        label = state_label(ps_states(p, i), spin=.true.)
        select case (len_trim(label))
        case (2)
          write(message(n),'(5X,A,5X,F7.2,7X,F8.3,6X,F8.3,10X,F8.3)') trim(label), &
               state_charge(ps_states(p, i)), &
               state_outermost_node(ps_states(p, i)), &
               state_outermost_peak(ps_states(p, i)), &
               state_default_rc(ps_states(p, i), ps_gen%scheme)
        case (5)
          write(message(n),'(4X,A,3X,F7.2,7XF8.3,6X,F8.3,10X,F8.3)') trim(label), &
               state_charge(ps_states(p, i)), &
               state_outermost_node(ps_states(p, i)), &
               state_outermost_peak(ps_states(p, i)), &
               state_default_rc(ps_states(p, i), ps_gen%scheme)
        end select
      end do
    end do

    if (present(unit)) then
      call write_info(n, unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(n, verbose_limit)
      else
        call write_info(n)
      end if
    end if

    call pop_sub()
  end subroutine ps_generation_output_input

end module ps_generation
