!! Copyright (C) 2004-2010 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module states
  use global
  use oct_parser
  use io
  use messages
  use utilities
  use units
  use splines
  use output
  use mesh
  use ps_io
  use quantum_numbers
  use potentials
  use wave_equations
  use eigensolver_m
  use wave_equations_integrator
  use hamann
  use troullier_martins
  use multireference
  implicit none


                    !---Interfaces---!

  interface state_density
    module procedure state_density_s, state_density_m
  end interface

  interface state_density_grad
    module procedure state_density_grad_s, state_density_grad_m
  end interface

  interface state_density_lapl
    module procedure state_density_lapl_s, state_density_lapl_m
  end interface

  interface state_tau
    module procedure state_tau_s, state_tau_m
  end interface

  interface state_density_moment
    module procedure state_density_moment_s, state_density_moment_m
  end interface

  interface assignment (=)
    module procedure state_copy
  end interface

  interface operator (==)
    module procedure state_equal
  end interface


                    !---Derived Data Types---!

  type state_type
    ! General information about the state
    type(qn_type)    :: qn    ! state quantum numbers
    real(R8)         :: occ   ! occupation
    real(R8)         :: ev    ! eigenvalue
    character(len=5) :: label ! a label to identify the state
    integer          :: wave_eq ! wave-equation used to obtain the wave-functions

    ! The wavefunctions
    integer :: np
    real(R8), pointer :: g(:) !  Schrodinger equation:
    real(R8), pointer :: f(:) !   g -> wavefunction
                              !   f -> wavefunction derivative
                              !  Dirac equation:
                              !   g -> spinor major component
                              !   f -> spinor minor component
                              !  Scalar-relativistic equation
                              !   g -> wavefunction
                              !   f -> wavefunction derivative

    ! Some information about the wavefunctions
    real(R8) :: peak ! outermost peak position
    real(R8) :: node ! outermost node position
  end type state_type


                    !---Global Variables---!

  integer, parameter :: NONE  = 0, &
                        HAM   = 1, &
                        TM    = 2, &
                        RTM   = 3, &
                        MRPP  = 4, &
                        RMRPP = 5, &
                        MTM   = 6


                    !---Public/Private Statements---!

  private
  public :: state_type, &
            state_null, &
            state_init, &
            state_end, &
            state_save, &
            state_load, &
            assignment(=), &
            operator(==), &
            state_sq, &
            state_density, &
            state_density_grad, &
            state_density_lapl, &
            state_tau, &            
            state_eigenvalue, &
            state_charge, &
            state_density_moment, &
            state_label, &
            state_qn, &
            state_update, &
            state_ld, &
            state_dipole_matrix_element, &
            states_output_configuration, &
            states_output_eigenvalues, &
            state_output_wf, &
            states_output_density, &
            state_outermost_peak, &
            state_outermost_node, &
            state_default_rc, &
            state_psp_generation, &
            state_test_consistency, &
            state_kb_projector, &
            state_test_ghost, &
            states_ps_io_set, &
            NONE, HAM, TM, RTM, MRPP, RMRPP, MTM

contains

  subroutine state_null(state)
    !-----------------------------------------------------------------------!
    !  Nullifies and sets to zero all the components of the state.        !
    !-----------------------------------------------------------------------!
    type(state_type), intent(out) :: state

    call push_sub("state_null")

    state%qn = QN_NULL
    state%ev = M_ZERO
    state%occ = M_ZERO
    state%label = ""
    state%wave_eq = 0
    state%peak = M_ZERO
    state%node = M_ZERO
    state%np = 0
    nullify(state%f)
    nullify(state%g)    

    call pop_sub()
  end subroutine state_null

  subroutine state_init(state, m, qn, occ, label)
    !-----------------------------------------------------------------------!
    ! Inititalizes a state. At this stage only the quantum numbers and the  !
    ! occupations are set. Wavefunctions are allocated and set to a         !
    ! constant.                                                             !
    !                                                                       !
    !  state - the state to be initialized                                  !
    !  m     - the mesh                                                     !
    !  qn    - the quantum numbers                                          !
    !  occ   - the occupation                                               !
    !-----------------------------------------------------------------------!
    type(state_type), intent(inout) :: state
    type(mesh_type),  intent(in)    :: m
    type(qn_type),    intent(in)    :: qn
    real(R8),         intent(in)    :: occ
    character(len=5), intent(in)    :: label

    call push_sub("state_init")

    state%qn = qn
    state%occ = occ
    state%label = label

    state%np = m%np
    allocate(state%f(m%np), state%g(m%np))
    state%g = M_ZERO
    state%f = M_ZERO

    call pop_sub()
  end subroutine state_init

  subroutine state_end(state)
    !-----------------------------------------------------------------------!
    ! Frees all memory associated to a state.                               !
    !-----------------------------------------------------------------------!
    type(state_type), intent(inout) :: state

    call push_sub("state_end")

    state%qn = QN_NULL
    state%ev = M_ZERO
    state%occ = M_ZERO
    state%label = ""
    state%wave_eq = 0
    state%peak = M_ZERO
    state%node = M_ZERO
    state%np = 0
    if (associated(state%g))  deallocate(state%g)
    if (associated(state%f))  deallocate(state%f)

    call pop_sub()
  end subroutine state_end

  subroutine state_copy(state_a, state_b)
    !-----------------------------------------------------------------------!
    ! Copies state_b to state_a.                                            !
    !-----------------------------------------------------------------------!
    type(state_type), intent(inout) :: state_a
    type(state_type), intent(in)    :: state_b

    call push_sub("state_copy")

    call state_end(state_a)

    state_a%qn = state_b%qn
    state_a%ev = state_b%ev
    state_a%occ = state_b%occ
    state_a%label = state_b%label
    state_a%wave_eq = state_b%wave_eq
    state_a%peak = state_b%peak
    state_a%node = state_b%node
    state_a%np = state_b%np
    allocate(state_a%g(state_b%np), state_a%f(state_b%np))
    state_a%g = state_b%g
    state_a%f = state_b%f

    call pop_sub()
  end subroutine state_copy

  subroutine state_save(unit, state)
    !-----------------------------------------------------------------------!
    ! Writes the state information to a file.                               !
    !                                                                       !
    !  unit  - file unit number                                             !
    !  state - state to be written                                          !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: unit
    type(state_type), intent(in) :: state

    integer :: n

    call push_sub("state_save")

    write(unit) state%qn
    write(unit) state%occ, state%ev, state%label, state%wave_eq, &
                state%np, state%peak, state%node

    do n = 1, state%np
      write(unit) state%g(n), state%f(n)
    end do

    call pop_sub()
  end subroutine state_save

  subroutine state_load(unit, state)
    !-----------------------------------------------------------------------!
    ! Reads the state information from a file.                              !
    !                                                                       !
    !  unit  - file unit number                                             !
    !  state - state to be read                                             !
    !-----------------------------------------------------------------------!
    integer,          intent(in)    :: unit
    type(state_type), intent(inout) :: state

    integer :: n

    call push_sub("state_load")

    read(unit) state%qn
    read(unit) state%occ, state%ev, state%label, state%wave_eq, &
               state%np, state%peak, state%node

    allocate(state%f(state%np), state%g(state%np))
    do n = 1, state%np
      read(unit) state%g(n), state%f(n)
    end do

    call pop_sub()
  end subroutine state_load

  elemental function state_equal(state_a, state_b)
    !-----------------------------------------------------------------------!
    ! Returns true if state_a and state_b have the same label.              !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state_a, state_b
    logical :: state_equal
    
    state_equal = state_a%label == state_b%label

  end function state_equal

  function state_sq(state)
    !-----------------------------------------------------------------------!
    ! Computes  the square of the wavefunction.                             !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_sq(state%np)

    call push_sub("state_sq")

    if (state%wave_eq /= DIRAC) then
      state_sq = state%g**2
    else
      state_sq = (state%g**2 + state%f**2)
    end if

    call pop_sub()
  end function state_sq

  function state_density_s(state)
    !-----------------------------------------------------------------------!
    ! Computes the electronic density associated with a state.              !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_density_s(state%np)

    call push_sub("state_density_s")

    state_density_s = state_sq(state)/(M_FOUR*M_PI)*state%occ

    call pop_sub()
  end function state_density_s

  function state_density_m(nspin, n_states, states, m)
    !-----------------------------------------------------------------------!
    ! Computes the electronic density associated with a set of states.      !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_density_m(m%np, nspin)

    integer :: p, i

    call push_sub("calculate_density_m")

    do p = 1, nspin
      state_density_m(:, p) = M_ZERO
      do i = 1, n_states
        state_density_m(:, p) = state_density_m(:, p) + state_density_s(states(p, i))
      end do
    end do

    call pop_sub()
  end function state_density_m

  function state_density_grad_s(state, m)
    !-----------------------------------------------------------------------!
    ! Computes the gradient of the electronic density associated with a     !
    ! state.                                                                !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_density_grad_s(state%np)

    call push_sub("state_density_grad_s")

    if (state%wave_eq /= DIRAC) then
      state_density_grad_s = M_TWO*state%g*spline_deriv(m%np, m%r, state%g, 3)
    else
      state_density_grad_s = M_TWO*(state%g*spline_deriv(m%np, m%r, state%g, 3) + &
                           state%f*spline_deriv(m%np, m%r, state%f, 3))
    end if
    
    state_density_grad_s = state_density_grad_s/(M_FOUR*M_PI)*state%occ

    call pop_sub()
  end function state_density_grad_s

  function state_density_grad_m(nspin, n_states, states, m)
    !-----------------------------------------------------------------------!
    ! Computes the gradient of the electronic density associated with a set !
    ! of states.                                                            !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_density_grad_m(m%np, nspin)

    integer :: p, i

    call push_sub("calculate_density_grad_m")

    do p = 1, nspin
      state_density_grad_m(:, p) = M_ZERO
      do i = 1, n_states
        state_density_grad_m(:, p) = state_density_grad_m(:, p) + &
                                            state_density_grad_s(states(p, i), m)
      end do
    end do

    call pop_sub()
  end function state_density_grad_m

  function state_density_lapl_s(state, m)
    !-----------------------------------------------------------------------!
    ! Computes the laplacian of the electronic density associated with a    !
    ! state.                                                                !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_density_lapl_s(state%np)

    call push_sub("state_density_lapl_s")

    if (state%wave_eq /= DIRAC) then
      state_density_lapl_s = &
                         M_TWO*(spline_deriv(m%np, m%r, state%g, 3)**2 + &
                         state%g*spline_deriv2(m%np, m%r, state%g, 3))
    else
      state_density_lapl_s = &
                         M_TWO*(spline_deriv(m%np, m%r, abs(state%g), 3)**2 + &
                                spline_deriv(m%np, m%r, abs(state%f), 3)**2 + &
                       abs(state%g)*spline_deriv2(m%np, m%r, abs(state%g), 3) + &
                       abs(state%f)*spline_deriv2(m%np, m%r, abs(state%f), 3))
    end if
    state_density_lapl_s = state_density_lapl_s/(M_FOUR*M_PI)*state%occ &
                           + M_TWO/m%r*state_density_grad_s(state, m)

    call pop_sub()
  end function state_density_lapl_s

  function state_density_lapl_m(nspin, n_states, states, m)
    !-----------------------------------------------------------------------!
    ! Computes the gradient of the electronic density associated with a set !
    ! of states.                                                            !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_density_lapl_m(m%np, nspin)

    integer :: p, i

    call push_sub("calculate_density_lapl_m")

    do p = 1, nspin
      state_density_lapl_m(:, p) = M_ZERO
      do i = 1, n_states
        state_density_lapl_m(:, p) = state_density_lapl_m(:, p) + &
                                            state_density_lapl_s(states(p, i), m)
      end do
    end do

    call pop_sub()
  end function state_density_lapl_m

  function state_tau_s(state, m)
    !-----------------------------------------------------------------------!
    ! Computes the kinetic energy density associated with a state.          !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_tau_s(m%np)

    integer :: i
    real(R8) :: l
    real(R8), allocatable :: df(:), dg(:)

    call push_sub("state_tau_s")

    if (state%occ == 0) then
      state_tau_s = M_ZERO
    else
      l = state%qn%l
      allocate(dg(m%np))
      dg = spline_deriv(m%np, m%r, state%g, 3)
        
      if (state%wave_eq /= DIRAC) then
        state_tau_s = dg**2 + l*(l + M_ONE)*state%g**2/m%r**2
      else
        allocate(df(m%np))
        df = spline_deriv(m%np, m%r, state%f, 3)

        state_tau_s = df**2 + dg**2 + &
                      M_TWO*(state%qn%k + M_ONE)/m%r*state%g*dg - &
                      M_TWO*(state%qn%k - M_ONE)/m%r*state%f*df + &
                      ((state%qn%k + M_ONE)/m%r*state%g)**2 + &
                      ((state%qn%k - M_ONE)/m%r*state%f)**2

        deallocate(df)
      end if
      deallocate(dg)
      state_tau_s = state_tau_s/(M_FOUR*M_PI)*state%occ
    end if

    call pop_sub()
  end function state_tau_s

  function state_tau_m(nspin, n_states, states, m)
    !-----------------------------------------------------------------------!
    ! Computes the kinetic energy density associated with a set of states.  !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    real(R8) :: state_tau_m(m%np, nspin)

    integer :: p, i

    call push_sub("calculate_tau_m")

    do p = 1, nspin
      state_tau_m(:, p) = M_ZERO
      do i = 1, n_states
        state_tau_m(:, p) = state_tau_m(:, p) + state_tau_s(states(p, i), m)
      end do
    end do

    call pop_sub()
  end function state_tau_m

  function state_density_moment_s(state, m, order)
    !-----------------------------------------------------------------------!
    ! Computes the density moment < R(r) | r**order | R(r)>.                !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(mesh_type),  intent(in) :: m
    integer,          intent(in) :: order
    real(R8) :: state_density_moment_s

    call push_sub("calculate_density_moment_s")

    if (state%wave_eq /= DIRAC) then
      state_density_moment_s = state%occ*spline_integ(m%np, m%r, state%g**2*m%r**(order+2), M_ZERO, m%r(m%np), 3)
    else
      state_density_moment_s = state%occ*spline_integ(m%np, m%r, (state%g**2 + state%f**2)*m%r**(order+2), M_ZERO, m%r(m%np), 3)
    end if

    call pop_sub()
  end function state_density_moment_s

  function state_density_moment_m(nspin, n_states, states, m, order)
    !-----------------------------------------------------------------------!
    ! Computes the density moment sum_i < R_i(r) | r**order | R_i(r)> of a  !
    ! given set of states.                                                  !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    integer,          intent(in) :: order
    real(R8) :: state_density_moment_m

    integer :: p, i

    call push_sub("calculate_density_moment_m")

    state_density_moment_m = M_ZERO
    do p = 1, nspin
      do i = 1, n_states
        state_density_moment_m = state_density_moment_m + state_density_moment_s(states(p, i), m, order)
      end do
    end do

  end function state_density_moment_m

  elemental function state_eigenvalue(state)
    !-----------------------------------------------------------------------!
    ! Returns the eigenvalue of a state.                                    !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_eigenvalue

    state_eigenvalue = state%ev

  end function state_eigenvalue

  elemental function state_charge(state)
    !-----------------------------------------------------------------------!
    ! Returns the number of electrons in an state.                          !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_charge

    state_charge = state%occ

  end function state_charge

  function state_label(state, spin)
    !-----------------------------------------------------------------------!
    ! Returns a label identifying the state. If "spin" is true the label    !
    ! includes information about the spin.                                  !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    logical,          intent(in), optional :: spin
    character(len=5) :: state_label

    logical :: spin_

    state_label = state%label

    spin_ = .false.
    if (present(spin)) spin_ = spin

    if (spin_ .and. state%qn%s /= M_ZERO) then
      if (state%qn%s == M_HALF) then
        state_label = trim(state%label)//"_up"
      elseif (state%qn%s == -M_HALF) then
        state_label = trim(state%label)//"_dn"
      end if
    end if

  end function state_label

  elemental function state_qn(state)
    !-----------------------------------------------------------------------!
    ! Returns the quantum numbers of the state.                             !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(qn_type) :: state_qn

    state_qn = state%qn

  end function state_qn

  subroutine state_update(state, m, wave_eq, potential, integrator_dp, &
       integrator_sp, eigensolver, ev)
    !-----------------------------------------------------------------------!
    ! Gets the states eigenvalue and eigenfunctions for a given potential.  !
    !                                                                       !
    !  state         - state                                                !
    !  m             - mesh                                                 !
    !  wave_eq       - wave-equation to solve                               !
    !  potential     - potential to use in the wave-equation                !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  eigensolver   - information about the eigensolver                    !
    !  ev            - energy                                               !
    !-----------------------------------------------------------------------!
    type(state_type),       intent(inout) :: state
    type(mesh_type),        intent(in)    :: m
    integer,                intent(in)    :: wave_eq
    type(potential_type),   intent(in)    :: potential
    type(integrator_type),  intent(inout) :: integrator_dp
    type(integrator_type),  intent(inout), optional :: integrator_sp
    type(eigensolver_type), intent(in),    optional :: eigensolver
    real(R8),               intent(in),    optional :: ev

    integer :: k

    call push_sub("state_update")

    if (present(ev)) then
      state%ev = ev
      
      ! Compute new wavefunctions
      call wavefunctions(state%qn, ev, wave_eq, m, potential, integrator_dp, &
                         state%g, state%f)
    else
      ! Get the new eigenvalues and eigenfunctions
      ASSERT(present(eigensolver) .and. present(integrator_sp))
      call eigensolve(state%qn, state%occ, wave_eq, m, potential, integrator_sp,&
                      integrator_dp, eigensolver, state%ev, state%g, state%f)
    end if
    state%wave_eq = wave_eq

    !Outermost peak
    if (state%ev /= M_ZERO) then
      do k = 1, m%np - 1
        if (abs(state%g(m%np-k+1)*m%r(m%np-k+1)) > abs(state%g(m%np-k)*m%r(m%np-k)) &
             .and. abs(state%g(m%np-k+1)*m%r(m%np-k+1)) > 1.0E-5_r8) then
          state%peak = m%r(m%np-k+1)
          exit
        end if
      end do
    else
      state%peak = M_ZERO
    end if

    !Outermost node
    if (state%ev /= M_ZERO) then
      do k = 1, m%np - 1
        if (state%g(m%np-k+1)*state%g(m%np-k) < M_ZERO) then
          state%node = m%r(m%np-k+1)
          exit
        end if
      end do
    else
      state%node = M_ZERO
    end if

    call pop_sub()
  end subroutine state_update

  function state_ld(state, e, r, wave_eq, integrator, potential, m)
    !-----------------------------------------------------------------------!
    ! Returns the logarithmic derivative of the wave-functions at r.        !
    !                                                                       !
    !  state      - state                                                   !
    !  e          - energy                                                  !
    !  r          - radius where to compute the logarithmic derivative      !
    !  wave_eq    - wave-equation to solve                                  !
    !  integrator - integrator object                                       !
    !  potential  - potential to use in the wave-equation                   !
    !  m          - mesh                                                    !
    !-----------------------------------------------------------------------!
    type(state_type),      intent(in)    :: state
    real(R8),              intent(in)    :: e, r
    integer,               intent(in)    :: wave_eq
    type(integrator_type), intent(inout) :: integrator
    type(potential_type),  intent(in)    :: potential
    type(mesh_type),       intent(in)    :: m
    real(R8) :: state_ld

    integer :: nnodes
    type(mesh_type) :: mr

    call push_sub("state_ld")

    !New mesh
    call mesh_null(mr)
    mr%np = locate(m%r, r, 0) + 1
    allocate(mr%r(mr%np))
    mr%r(1:mr%np-1) = m%r(1:mr%np-1)
    mr%r(mr%np) = r

    !Log derivative a la Hamann
    state_ld = hamann_ld(state%qn, e, wave_eq, mr, potential, integrator, nnodes)

    call mesh_end(mr)

    call pop_sub()
  end function state_ld

  function state_dipole_matrix_element(m, state_i, state_f)
    !-----------------------------------------------------------------------!
    ! Computes the dipole matrix element between two states                 !
    !                                                                       !
    !                     M_if = < R_i(r) | r | R_f(r) >                    !
    !                                                                       !
    ! Notice that the angular part is not taken into  account. This is OK,  !
    ! because of the spatial averaging and spherical symmetry.              !
    !                                                                       !
    !  m       - mesh                                                       !
    !  state_i - state i                                                    !
    !  state_f - state f                                                    !
    !-----------------------------------------------------------------------!
    type(mesh_type),  intent(in) :: m
    type(state_type), intent(in) :: state_i, state_f
    real(R8) :: state_dipole_matrix_element

    real(R8), allocatable :: tmp(:)

    call push_sub("state_dipole_matrix_element")
    
    if (state_i%qn%s /= state_i%qn%s .or. &
         (state_i%qn%n == state_f%qn%n .and. state_i%qn%l == state_f%qn%l .and. &
         state_i%qn%j /= state_f%qn%j) ) then
      state_dipole_matrix_element = M_ZERO
      return
    end if

    allocate(tmp(m%np))

    if (state_i%wave_eq /= DIRAC .and. state_f%qn%j /= DIRAC) then
      tmp = (state_i%g*state_f%g)*m%r**3
    elseif (state_i%wave_eq == DIRAC .and. state_f%wave_eq == DIRAC) then
      tmp = (state_i%g*state_f%g + state_i%f*state_f%f)*m%r**3
    else
      message(1) = "Incompatible wavefunctions in routine state_dipole_matrix_element"
      call write_fatal(1)
    end if

    state_dipole_matrix_element = spline_integ(m%np, m%r, tmp, M_ZERO, m%r(m%np), 3)

    deallocate(tmp)

    call pop_sub()
  end function state_dipole_matrix_element

  function state_outermost_peak(state)
    !-----------------------------------------------------------------------!
    ! Returns the positions of the wave-functions outermost peak.           !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_outermost_peak

    call push_sub("state_outermost_peak")

    state_outermost_peak = state%peak

    call pop_sub()
  end function state_outermost_peak

  function state_outermost_node(state)
    !-----------------------------------------------------------------------!
    ! Returns the positions of the wave-functions outermost node.           !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    real(R8) :: state_outermost_node

    call push_sub("state_outermost_node")

    state_outermost_node = state%node

    call pop_sub()
  end function state_outermost_node

  elemental function state_default_rc(state, scheme)
    !-----------------------------------------------------------------------!
    ! Returns the default cut-off radius.
    !-----------------------------------------------------------------------!
    type(state_type), intent(in)  :: state
    integer,          intent(in)  :: scheme
    real(R8) :: state_default_rc

    !Get default core radius
    select case (scheme)
    case (HAM)
      if (state%qn%n > state%qn%l + 1) then
        state_default_rc = 0.6_r8*state%peak
      else
        state_default_rc = 0.4_r8*state%peak
      end if
    case (TM)
      state_default_rc = M_ZERO
    case (RTM)
      state_default_rc = M_ZERO
    case (MRPP)
      state_default_rc = M_ZERO
    case (RMRPP)
      state_default_rc = M_ZERO
    case (MTM)
      state_default_rc = M_ZERO
    end select

  end function state_default_rc

  subroutine state_psp_generation(m, scheme, wave_eq, tol, ae_potential, &
                                  integrator_sp, integrator_dp, ps_v, state1, &
                                  rc1, state2, rc2)
    !-----------------------------------------------------------------------!
    ! Generate the pseudo wave-functions and the pseudo-potential using     !
    ! one of the implemented schemes.                                       !
    !                                                                       !
    !  m             - mesh                                                 !
    !  scheme        - scheme used to generate the pseudo-potentials        !
    !  wave_eq       - wave-equation to use                                 !
    !  tol           - tolerance                                            !
    !  ae_potential  - all-electron potential                               !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  ps_v          - pseudo-potential on the mesh                         !
    !  state1        - pseudo-state 1 (either valence or semi-core)         !
    !  rc1           - core radius of state 1                               !
    !  state2        - valence state when state 1 is a semi-core state      !
    !  rc2           - core radius of state 2                               !
    !-----------------------------------------------------------------------!
    type(mesh_type),       intent(in)    :: m
    integer,               intent(in)    :: scheme, wave_eq
    real(R8),              intent(in)    :: tol
    type(potential_type),  intent(in)    :: ae_potential
    type(integrator_type), intent(inout) :: integrator_sp, integrator_dp
    real(R8),              intent(out)   :: ps_v(m%np)
    type(state_type),      intent(inout) :: state1
    real(R8),              intent(in)    :: rc1
    type(state_type),      intent(inout), optional :: state2
    real(R8),              intent(in),    optional :: rc2

    integer :: scheme_
    character(len=5) :: label

    call push_sub("state_psp_generation")

    scheme_ = scheme
    if (present(state2)) then
      ASSERT(present(rc2))
    else
      select case (scheme_)
      case (MRPP)
        scheme_ = TM
      case (RMRPP)
        scheme_ = RTM
      end select
    end if

    !Write information
    label = state_label(state1, spin=.true.)
    write(message(1),'(2x,"State: ",a,a)') trim(label)
    if (present(state2)) then
      label = state_label(state2, spin=.true.)
      write(message(1),'(a,2x,a)') trim(message(1)), trim(label)
    end if
    call write_info(1,20)
    call write_info(1,unit=out_pp)

    select case (scheme_)
    case (HAM) !Hamann scheme
      call hamann_gen(state1%qn, state1%ev, wave_eq, tol, m, ae_potential, &
                      rc1, integrator_sp, integrator_dp, ps_v, &
                      state1%g, state1%f)

    case (TM) !Troullier-Martins scheme
      if (present(state2)) then
        !Get valence wave-function first
        call tm_gen(state2%qn, state2%ev, wave_eq, .false., tol, m, &
                    ae_potential, rc2, ps_v, state2%g, state2%f)
        state2%qn%n = state2%qn%n + 1
      end if

      call tm_gen(state1%qn, state1%ev, wave_eq, .false., tol, m, &
                  ae_potential, rc1, ps_v, state1%g, state1%f)
      
    case (RTM) !Relativistic extension of the Troullier-Martins scheme
      if (present(state2)) then
        !Get valence wave-function first
        call tm_gen(state2%qn, state2%ev, wave_eq, .true., tol, m, &
                    ae_potential, rc2, ps_v, state2%g, state2%f)
        state2%qn%n = state2%qn%n + 1
      end if

      call tm_gen(state1%qn, state1%ev, wave_eq, .true., tol, m, &
                  ae_potential, rc1, ps_v, state1%g, state1%f)

    case (MRPP) !Multireference Pseudopotentials
      call mrpp_gen(state1%qn, state2%qn, state1%ev, state2%ev, &
                    wave_eq, .false., tol, m, ae_potential, rc1, &
                    rc2, integrator_dp, ps_v, state1%g, state1%f, &
                    state2%g, state2%f)

    case (RMRPP) !Multireference Pseudopotentials
      call mrpp_gen(state1%qn, state2%qn, state1%ev, state2%ev, &
                    wave_eq, .true., tol, m, ae_potential, rc1, &
                    rc2, integrator_dp, ps_v, state1%g, state1%f, &
                    state2%g, state2%f)

    end select

    call pop_sub()
  end subroutine state_psp_generation

  subroutine state_test_consistency(m, wave_eq, eigensolver, integrator_sp, &
                          integrator_dp, ae_potential, ps_potential, ps_state, &
                          rmatch, ev, norm, slope)
    !-----------------------------------------------------------------------!
    ! Test the consistency of the pseudo-potential and pseudo               !
    ! wave-functions. The test consists in solving the wave-equations for   !
    ! the pseudo-potential and then comparing the resulting eigenvalue and  !
    ! wave-functions with the all-electron ones.                            !
    !                                                                       !
    !  m             - mesh                                                 !
    !  wave_eq       - wave-equation to use                                 !
    !  eigensolver   - information about the eigensolver                    !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  ae_potential  - all-electron potential                               !
    !  ps_potential  - pseudo-potential                                     !
    !  ps_state      - pseudo-state                                         !
    !  rmatch        - matching radius                                      !
    !  ev            - eigenvalue                                           !
    !  norm          - norm of the wave-function up to rc                   !
    !  slope         - slope of the wave-functions at rc                    !  
    !-----------------------------------------------------------------------!
    type(mesh_type),        intent(in)    :: m
    integer,                intent(in)    :: wave_eq
    type(eigensolver_type), intent(in)    :: eigensolver
    real(R8),               intent(in)    :: rmatch
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(potential_type),   intent(in)    :: ps_potential, ae_potential
    type(state_type),       intent(in)    :: ps_state
    real(R8),               intent(out)   :: ev, norm, slope

    integer :: i
    real(R8), allocatable  :: tmp(:)
    type(state_type) :: state
    type(spline_type) :: spl

    call push_sub("state_test_consistency")

    call spline_null(spl)
    call state_null(state)

    state = ps_state
    call state_update(state, m, wave_eq, ps_potential, integrator_dp, &
                      integrator_sp, eigensolver)
    ev = state%ev

    allocate(tmp(m%np))
    call spline_init(spl, m%np, m%r, state%g, 3)
    do i = 1, m%np
      tmp(i) = (spline_eval(spl, m%r(i))*m%r(i))**2
    end do
    norm = spline_integ(m%np, m%r, tmp, M_ZERO, rmatch, 3)
    slope = spline_eval_deriv(spl, rmatch)
    call spline_end(spl)

    call state_update(state, m, wave_eq, ae_potential, integrator_dp, ev=ev)
    call spline_init(spl, m%np, m%r, state%g, 3)
    do i = 1, m%np
      tmp(i) = (spline_eval(spl, m%r(i))*m%r(i))**2
    end do
    norm = norm/spline_integ(m%np, m%r, tmp, M_ZERO, rmatch, 3)
    slope = abs(slope/spline_eval_deriv(spl, rmatch))
    call spline_end(spl)
    call state_end(state)
    deallocate(tmp)

    call pop_sub()
  end subroutine state_test_consistency

  subroutine state_kb_projector(m, nspin, local_ps_potential, ps_potential, &
                                state, qn, e, proj_f)
    !-----------------------------------------------------------------------!
    ! Computes the KB projector defined in the following way:               !
    !                                                                       !
    ! KB projector = | proj_f > E < proj_f |                                !
    !                                                                       !
    !     <phi_l (v_l - v_local) | (v_l - v_local) phi_l>                   !
    ! E = -----------------------------------------------                   !
    !          < phi_l | (v_l - v_local) | phi_l >                          !
    !                                                                       !
    !                             | (v_l - v_local) phi_l>                  !
    ! | proj_f > = -------------------------------------------------------  !
    !              || < phi_l (v_l - v_local) | (v_l - v_local) phi_l > ||  !
    !                                                                       !
    !                                                                       !
    !  m                  - mesh                                            !
    !  nspin              - number of spin channels                         !
    !  local_ps_potential - local component of the KB form                  !
    !  ps_potential       - pseudo-potential                                !
    !  state              - pseudo-state                                    !
    !  qn                 - the quantum numbers of the pseudo-state         !
    !  e                  - KB energy                                       !
    !  proj_f             - KB projector function                           !
    !-----------------------------------------------------------------------!
    type(mesh_type),      intent(in)  :: m
    integer,              intent(in)  :: nspin
    type(potential_type), intent(in)  :: local_ps_potential, ps_potential
    type(state_type),     intent(in)  :: state(nspin)
    type(qn_type),        intent(out) :: qn
    real(R8),             intent(out) :: e
    real(R8),             intent(out) :: proj_f(m%np)

    integer :: i
    real(R8) :: int1, int2, occ_sum, cos
    real(R8), parameter :: delta = 1.0e-20_r8
    real(R8), allocatable :: vphi(:), phi(:)
    character(len=5) :: label

    call push_sub("state_kb_projector")

    allocate(vphi(m%np), phi(m%np))

    if (nspin == 1) then
      phi = state(1)%g
    else
      occ_sum = sum(state%occ)
      if (occ_sum == M_ZERO) then
        phi = state(1)%g*M_HALF + state(2)%g*M_HALF
      else
        phi = (state(1)%g*state(1)%occ + state(2)%g*state(2)%occ)/occ_sum
      end if
      phi = phi/sqrt(spline_integ(m%np, m%r, phi**2*m%r**2, M_ZERO, m%r(m%np), 3))
    end if

    qn = state(1)%qn
    do i = 1, m%np
      vphi(i) = (v(ps_potential, m%r(i), qn, unscreened=.true.) &
           - v(local_ps_potential, m%r(i), qn, unscreened=.true.))*phi(i)
    end do
    int1 = spline_integ(m%np, m%r, vphi**2*m%r**2, M_ZERO, m%r(m%np), 3)
    int2 = spline_integ(m%np, m%r, vphi*phi*m%r**2,  M_ZERO, m%r(m%np), 3)

    if (int1 /= M_ZERO) then
      e = int1/(int2 + delta)
      proj_f = vphi/sqrt(int1)
      cos = sqrt(int1)/e
    else
      e = M_ZERO
      proj_f = M_ZERO
      cos = M_ZERO
    end if

    if (abs(e) < 10e-12_r8) then
      e = M_ZERO
      proj_f = M_ZERO
      cos = M_ZERO
    end if

    deallocate(vphi, phi)

    !Output information
    if (e /= M_ZERO) then
      label = state_label(state(1))
      select case (len_trim(label))
      case (2)
        write(message(1),'(6X,A,5X,F9.4,8X,F7.4)') trim(label), e/units_out%energy%factor, cos
      case (5)
        write(message(1),'(4X,A,4X,F9.4,8X,F7.4)') trim(label), e/units_out%energy%factor, cos
      end select
      call write_info(1)
      call write_info(1,unit=out_kb)
    end if

    call pop_sub()
  end subroutine state_kb_projector

  subroutine state_test_ghost(m, wave_eq, integrator_sp, integrator_dp, &
                              eigensolver, kb_potential, state)
    !-----------------------------------------------------------------------!
    ! Tests the KB form of the pseudo-potential for ghost states.           !
    !                                                                       !
    !  m             - mesh                                                 !
    !  wave_eq       - wave-equation to use                                 !
    !  integrator_sp - single-precision integrator object                   !
    !  integrator_dp - double-precision integrator object                   !
    !  eigensolver   - information about the eigensolver                    !
    !  kb_potential  - the KB potential                                     !
    !  state         - pseudo-state                                         !
    !-----------------------------------------------------------------------!
    type(mesh_type),        intent(in)    :: m
    integer,                intent(in)    :: wave_eq
    type(integrator_type),  intent(inout) :: integrator_sp, integrator_dp
    type(eigensolver_type), intent(in)    :: eigensolver
    type(potential_type),   intent(in)    :: kb_potential
    type(state_type),       intent(in)    :: state

    real(R8) :: kb_e, eloc0, eloc1, u
    character(len=5) :: label
    type(qn_type) :: qn

    call push_sub("state_test_ghost")

    kb_e = potential_kb_energy(kb_potential, state%qn)

    if (kb_e == M_ZERO) then
      call pop_sub()
      return
    end if

    qn = state%qn
    call eigensolve(qn, M_ZERO, wave_eq, m, kb_potential, integrator_sp, &
                    integrator_dp, eigensolver, eloc0)

    qn%n = qn%n + 1
    call eigensolve(qn, M_ZERO, wave_eq, m, kb_potential, integrator_sp, &
                    integrator_dp, eigensolver, eloc1)

    label = state_label(state, spin=.true.)
    u = units_out%energy%factor
    write(message(1),'(4x,"State: ",A)') trim(label)
    write(message(3),'(6X,"Local potential eigenvalues: ",F9.4," (E0)  ",F9.4," (E1)")') eloc0/u, eloc1/u
    write(message(4),'(6X,"Reference energy:            ",F9.4," (Eref)")') state%ev/u

    if (kb_e > M_ZERO) then
      if (state%ev > eloc0) then
        if (state%ev < eloc1) then
          write(message(2),'(6X,"KB energy > 0; E0 < Eref < E1  =>  No ghost states")')
        else
          if (eloc1 < M_ZERO) then
            write(message(2),'(6X,"KB energy > 0; Eref > E1       =>  Ghost state found")')
          else
            write(message(2),'(6X,"KB energy > 0; Eref = E1 = 0   =>  Unable to determine")')
          end if
        end if
      else
        write(message(2),'(6X,"KB energy > 0;  Eref < E0      =>  Illdefined")')
      end if
    elseif (kb_e < M_ZERO) then
      if (state%ev < eloc0) then
        write(message(2),'(6X,"KB energy < 0; Eref < E0       =>  No ghost states")')
      else
        if (eloc0 < M_ZERO) then
          write(message(2),'(6X,"KB energy < 0; Eref > E0       =>  Ghost state found")')
        else
          write(message(2),'(6X,"KB energy < 0; Eref = E0 = 0   =>  Unable to determine")')
        end if
      end if
    end if
    call write_info(4)
    call write_info(4, unit=out_kb)

    call pop_sub()
  end subroutine state_test_ghost

  subroutine states_output_configuration(states, nspin, n_states, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Outputs the states label and occupancies in a nice readable format.   !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(:,:)
    integer,          intent(in), optional :: unit, verbose_limit

    integer           :: i
 
    call push_sub("states_output_configuration")

    select case (nspin)
    case (1) !Spin-unpolarized
      write(message(1),'(2X,"Configuration    : State   Occupation")')
    case (2) !Spin-polarized
      write(message(1),'(2X,"Configuration    : State   Occupation   |  State   Occupation")')
    end select

    do i = 1, n_states
      if (states(1,i)%qn%j == M_ZERO) then
        write(message(i+1),'(23X,A,6X,F5.2)') trim(states(1,i)%label), states(1,i)%occ
      else
        write(message(i+1),'(21X,A,5X,F5.2)') trim(states(1,i)%label), states(1,i)%occ
      end if
    end do

    if (nspin == 2) then
      do i = 1, n_states
        if (states(2,i)%qn%j == M_ZERO) then
          write(message(i+1),'(A,"      |    ",A,6X,F5.2)') trim(message(i+1)), trim(states(2,i)%label), states(2,i)%occ
        else
          write(message(i+1),'(A,"      |  ",A,4X,F5.2)') trim(message(i+1)), trim(states(2,i)%label), states(2,i)%occ
        end if
      end do
    end if

    if (present(unit)) then
      call write_info(n_states+1, unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(n_states+1, verbose_limit)
      else
        call write_info(n_states+1)
      end if
    end if

    call pop_sub()
  end subroutine states_output_configuration

  subroutine states_output_eigenvalues(nspin, n_states, states, unocc, unit, verbose_limit)
    !-----------------------------------------------------------------------!
    ! Writes the eigenvalues either to a file or to the screen in a nice    !
    ! readable format.                                                      !
    !                                                                       !
    !  nspin    - number of spin channels                                   !
    !  n_states - number of states                                          !
    !  states   - the states                                                !
    !  unocc    - if true print the eigenvalues of unoccupied states        !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    logical,          intent(in) :: unocc
    integer,          intent(in), optional :: unit, verbose_limit

    logical  :: mask(n_states)
    integer  :: i, j, jd, p, dum(1), order(nspin, n_states)
    real(R8) :: u

    call push_sub("states_output_eigenvalues")

    !sort states
    do p = 1, nspin
      mask = .true.
      do i = 1, n_states
        dum = maxloc(abs(states(p, :)%ev), mask=mask)
        order(p, i) = dum(1)
        mask(order(p, i)) = .false.
      end do
    end do

    write(message(1),'(4X,"Eigenvalues [",A,"]")') trim(units_out%energy%abbrev)
    select case (nspin)
    case (1)
      write(message(2),'(5X,"State   Occupation   Eigenvalue")')
    case (2)
      write(message(2),'(5X,"State   Occupation   Eigenvalue  |  State   Occupation   Eigenvalue")')
    end select

    u = units_out%energy%factor
    i = 0; j = 2
    do
      if (i == n_states) exit
      i = i + 1
      if (states(1, order(1, i))%occ == 0 .and. .not. unocc) cycle
      j = j + 1
      if (states(1, order(1, i))%qn%j == M_ZERO) then
        write(message(j),'(7X,A,6X,F5.2,4X,F12.5)') trim(states(1, order(1, i))%label), states(1, order(1, i))%occ, &
                                                    states(1, order(1, i))%ev/u
      else
        write(message(j),'(5X,A,5X,F5.2,4X,F12.5)') trim(states(1, order(1, i))%label), states(1, order(1, i))%occ, &
                                                    states(1, order(1, i))%ev/u
      end if
    end do

    if (nspin == 2) then
      i = 0
      jd = 2
      do
        if (i == n_states) exit
        i = i + 1
        if (states(2, order(2, i))%occ == 0 .and. .not.unocc) cycle
        jd = jd + 1
        if (jd <= j) then
          if (states(2, order(2, i))%qn%j == M_ZERO) then
            write(message(jd),'(A,"  |",4X,A,6X,F5.2,4X,F12.5)') trim(message(jd)), trim(states(2, order(2, i))%label), &
                                                                 states(2, order(2, i))%occ, states(2,order(2, i))%ev/u
          else
            write(message(jd),'(A,"  |",2X,A,5X,F5.2,4X,F12.5)') trim(message(jd)), trim(states(2, order(2, i))%label), &
                                                                 states(2, order(2, i))%occ, states(2,order(2, i))%ev/u
          end if
        else
          if (states(2, order(2, i))%qn%j == M_ZERO) then
            write(message(jd),'(38X,"|",4X,A,6X,F5.2,4X,F12.5)') trim(states(2, order(2,i))%label), states(2, order(2,i))%occ, &
                                                                 states(2, order(2, i))%ev/u
          else
            write(message(jd),'(38X,"|",2X,A,5X,F5.2,4X,F12.5)') trim(states(2, order(2,i))%label), states(2, order(2,i))%occ, &
                                                                 states(2, order(2, i))%ev/u
          end if
        end if

      end do

    else
      jd = 0
    end if

    j = max(j,jd)
    if (present(unit)) then
      call write_info(j,unit=unit)
    else
      if (present(verbose_limit)) then
        call write_info(j,verbose_limit)
      else
        call write_info(j)
      end if
    end if

    call pop_sub()
  end subroutine states_output_eigenvalues

  subroutine state_output_wf(state, m, dir)
    !-----------------------------------------------------------------------!
    ! Writes the wave-function to a file in a format suitable for plotting. !
    !-----------------------------------------------------------------------!
    type(state_type), intent(in) :: state
    type(mesh_type),  intent(in) :: m
    character(len=*), intent(in) :: dir

    integer           :: i, unit, iostat
    real(R8)          :: u
    character(len=5)  :: label
    character(len=30) :: filename

    call push_sub("state_output_wf")

    call io_assign(unit)
    label = state_label(state, spin=.true.)
    filename = trim(dir)//"/wf-"//trim(label)
    open(unit, file=trim(filename), status='unknown', iostat=iostat)
    if (iostat > 0) then
      write(message(1),'("Unable to open file ''",A,"''")') trim(filename)
      call write_fatal(1)
    end if
    if (state%qn%j == M_ZERO) then
      write(unit,'("# Radial wavefunction and first derivative.")')
    else
      write(unit,'("# Radial wavefunctions: major and minor components.")')
    end if
    write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
    write(unit,'("# Length units: ",A)') trim(units_out%length%name)
    write(unit,'("#")')
    select case (int(2*state%qn%s))
    case (0)
      write(unit,'("# State: ",A)') trim(state%label)
    case (-1)
      write(unit,'("# State: ",A," down")') trim(state%label)
    case (1)
      write(unit,'("# State: ",A," up")') trim(state%label)
    end select

    u = units_out%length%factor

    write(unit,'("# Nodes: ")',advance='no')
    do i = 2, m%np
      if (state%g(i)*state%g(i-1) < M_ZERO) &
        write(unit,'(F12.6,2X)',advance='no') (m%r(i)+m%r(i-1))/(M_TWO*u)
    end do
    write(unit,*)
    write(unit,'("# Peaks: ")',advance='no')
    do i = 2, m%np-1
      if ( ( abs(state%g(i)*m%r(i)) > abs(state%g(i-1)*m%r(i-1)) ) &
            .and. ( abs(state%g(i)*m%r(i)) > abs(state%g(i+1)*m%r(i+1)) ) ) &
            write(unit,'(F12.6,2X)',advance='no') m%r(i)/u
    end do
    write(unit,*)

    do i = 1, m%np
      write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2)') m%r(i)/u, state%g(i)*sqrt(u), state%f(i)*sqrt(u)
    end do
    close(unit)

    call pop_sub()
  end subroutine state_output_wf

  subroutine states_output_density(nspin, n_states, states, m, dir)
    !-----------------------------------------------------------------------!
    ! Writes the electronic density to the "dir/density" file in a format   !
    ! suitable for plotting.                                                !
    !-----------------------------------------------------------------------!
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    type(mesh_type),  intent(in) :: m
    character(len=*), intent(in) :: dir

    integer  :: i, p, unit, iostat
    real(R8) :: u
    character(len=30) :: filename
    real(R8), allocatable :: rho(:,:), rho_grad(:,:), rho_lapl(:,:), tau(:,:)

    call push_sub("states_output_density")

    !Get the density and its derivatives
    allocate(rho(m%np, nspin), rho_grad(m%np, nspin))
    allocate(rho_lapl(m%np, nspin), tau(m%np, nspin))
    rho      = state_density     (nspin, n_states, states, m)
    rho_grad = state_density_grad(nspin, n_states, states, m)
    rho_lapl = state_density_lapl(nspin, n_states, states, m)
    tau      = state_tau         (nspin, n_states, states, m)

    call io_assign(unit)
    u = units_out%length%factor

    !Density, gradient, and laplacian
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/density"
      else
        if (p == 1) then
          filename = trim(dir)//"/density_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/density_up"
        end if
      end if
      open(unit, file=trim(filename), status='unknown', iostat=iostat)
      if (iostat > 0) then
        write(message(1),'("Unable to open file ''",A,"''")') trim(filename)
        call write_fatal(1)
      end if

      write(unit,'("#")')
      write(unit,'("# Radial density.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("#")')

      write(unit,'("# ",71("-"))')
      write(unit,'("# |",7X,"r",7X,"|",6X,"n(r)",7X,"|",4X,"grad_n(r)",4X,"|",4X,"lapl_n(r)",4X,"|")')
      write(unit,'("# ",71("-"))')
      do i = 1, m%np
        write(unit,'(3X,ES14.8E2,3X,ES15.8E2,3X,ES15.8E2,3X,ES15.8E2)') &
             m%r(i)/u, rho(i, p)*u, rho_grad(i, p)*u**2, rho_lapl(i, p)*u**3
      end do

      close(unit)
    end do

    !Print kinetic energy density
    do p = 1, nspin
      if (nspin == 1) then
        filename = trim(dir)//"/tau"
      else
        if (p == 1) then
          filename = trim(dir)//"/tau_dn"
        elseif (p == 2) then
          filename = trim(dir)//"/tau_up"
        end if
      end if
      open(unit, file=trim(filename), status='unknown', iostat=iostat)
      if (iostat > 0) then
        write(message(1),'("Unable to open file ''",A,"''")') trim(filename)
        call write_fatal(1)
      end if

      write(unit,'("#")')
      write(unit,'("# Radian kinetic energy density.")')
      write(unit,'("# Length units: ",A)') trim(units_out%length%name)
      write(unit,'("# Energy units: ",A)') trim(units_out%energy%name)
      write(unit,'("#")')

      write(unit,'("# ",35("-"))')
      write(unit,'("# |",7X,"r",7X,"|",6X,"n(r)",7X,"|")')
      write(unit,'("# ",35("-"))')
      do i = 1, m%np
        write(unit,'(3X,ES14.8E2,3X,ES15.8E2)') m%r(i)/u, &
                                              tau(i, p)*u/units_out%energy%factor
      end do

      close(unit)
    end do

    !Free memory
    deallocate(rho, rho_grad, rho_lapl, tau)

    call pop_sub()
  end subroutine states_output_density

  subroutine states_ps_io_set(m, nspin, n_states, states, rc)
    !-----------------------------------------------------------------------!
    ! Pass the information about the wavefunctions to the ps_io module.     !
    !-----------------------------------------------------------------------!
    type(mesh_type),  intent(in) :: m
    integer,          intent(in) :: nspin, n_states
    type(state_type), intent(in) :: states(nspin, n_states)
    real(R8),         intent(in) :: rc(n_states)

    integer :: i
    integer, allocatable :: n(:)
    real(R8), allocatable :: occ(:,:), wfs(:,:), rho_val(:,:)

    call push_sub("states_ps_io_set")

    allocate(n(n_states))
    allocate(wfs(m%np, n_states))
    allocate(occ(n_states, nspin))
    allocate(rho_val(m%np, nspin))
    rho_val = state_density(nspin, n_states, states, m)
    do i = 1, n_states
      read(states(1, i)%label,'(I1)') n(i)
      wfs(:, i) = states(1, i)%g
      occ(i, :) = states(:, i)%occ
    end do

    call ps_io_set_wfs(m%np, nspin, n_states, n, states(1, :)%qn%l, &
        states(1, :)%qn%j, occ, states(1, :)%ev, rc(:), wfs, sum(rho_val, dim=2))

    deallocate(n, wfs, occ, rho_val)

    call pop_sub()
  end subroutine states_ps_io_set

end module states
