!! Copyright (C) 2004-2009 M. Oliveira, F. Nogueira
!!
!! This program is free software; you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation; either version 2, or (at your option)
!! any later version.
!!
!! This program is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with this program; if not, write to the Free Software
!! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
!! 02111-1307, USA.

#include "global.h"

module troullier_martins
  use global
  use gsl_interface
  use messages
  use io
  use utilities
  use output
  use mesh
  use multiroot
  use splines
  use quantum_numbers
  use potentials
  use wave_equations_integrator
  use wave_equations
  use troullier_martins_equations
  implicit none


                    !---Public/Private Statements---!

  private
  public :: tm_gen, &
            tm_equations_rhs, &
            tm_solve_system


contains

  subroutine tm_gen(qn, ev, wave_eq, rel, tol, m, ae_potential, rc, ps_v, ps_g, ps_f)
    !-----------------------------------------------------------------------!
    ! Generates the pseudo wave-functions and the corresponding             !
    !pseudo-potential using the Troullier-Martins scheme.                   !
    !                                                                       !
    !  qn           - quantum numbers of valence state                      !
    !  occ          - occupation of valence state                           !
    !  ev           - eigenvalue of valence state                           !
    !  wave_eq      - wave equation to use                                  !
    !  rel          - if true, use the relativistic extension to the scheme !
    !  tol          - tolerance                                             !
    !  m            - mesh                                                  !
    !  ae_potential - all-electron potential                                !
    !  rc           - cutoff radius                                         !
    !  integrator   - integrator object                                     !
    !  ps_v         - pseudo-potential on the mesh                          !
    !  ps_g         - all-electron/pseudo wave-function                     !
    !  ps_f         - all-electron/pseudo wave-function                     !
    !                                                                       !
    ! On input qn, ps_f and ps_g contain the all-electron quantum-numbers   !
    ! and wavefunctions and on exit the pseudo wavefunctions and quantum    !
    ! numbers.                                                              !
    !-----------------------------------------------------------------------!
    type(qn_type),         intent(inout) :: qn
    real(R8),              intent(in)    :: ev
    integer,               intent(in)    :: wave_eq
    logical,               intent(in)    :: rel
    real(R8),              intent(in)    :: tol
    type(mesh_type),       intent(in)    :: m
    type(potential_type),  intent(in)    :: ae_potential
    real(R8),              intent(in)    :: rc
    real(R8),              intent(out)   :: ps_v(m%np)
    real(R8),              intent(inout) :: ps_g(m%np), ps_f(m%np)

    integer :: i
    real(R8) :: rhs(7), c(7)
    real(R8) :: grc, gprc, norm
    real(R8), allocatable :: ae_g(:), ae_gp(:), ae_f(:)
    type(mesh_type) :: mr
    type(spline_type) :: spl

    call push_sub("tm_gen")

    !
    call mesh_null(mr)
    call spline_null(spl)

    !Write core radius
    write(message(1),'(4x,"Core radius:",1x,f7.3,5x)') rc
    call write_info(1,20)
    call write_info(1,unit=out_pp)

    !Copy all-electron wavefunctions
    allocate(ae_g(m%np), ae_f(m%np), ae_gp(m%np))
    ae_g = ps_g
    ae_f = ps_f
    select case (wave_eq)
    case (SCHRODINGER, SCALAR_REL)
      ae_gp = ae_f
      ae_f  = M_ZERO
    case (DIRAC)
      ae_gp = spline_deriv(m%np, m%r, ae_g, 3)
    end select

    !Wavefunctions and norm at rc
    call spline_init(spl, m%np, m%r, ae_g, 3)
    grc  = rc*spline_eval(spl, rc)
    call spline_end(spl)
    call spline_init(spl, m%np, m%r, ae_gp, 3)
    gprc = rc*spline_eval(spl, rc) + grc/rc
    call spline_end(spl)
    if (grc < M_ZERO) then
      ae_g  = -ae_g
      ae_gp = -ae_gp
      ae_f  = -ae_f
      grc   = -grc
      gprc  = -gprc
    end if
    if (rel) then
      norm = log(spline_integ(m%np, m%r, m%r**2*(ae_g**2 + ae_f**2), M_ZERO, rc, 3))
    else
      select case (wave_eq)
      case (SCHRODINGER, SCALAR_REL)
        norm = log(spline_integ(m%np, m%r, (m%r*ae_g)**2, M_ZERO, rc, 3))
      case (DIRAC)
        norm = log(spline_integ(m%np, m%r, (m%r*ae_g)**2, M_ZERO, rc, 3) + &
               spline_integ(m%np, m%r, (m%r*ae_f)**2, M_ZERO, m%r(m%np), 3))
      end select
    end if

    !Get right hand side of equations to solve
    call tm_equations_rhs(m, rc, rel, qn, ev, grc, gprc, norm, ae_potential, rhs)

    !New mesh
    mr%type = m%type
    mr%np = locate(m%r, rc, 0) + 1
    allocate(mr%r(mr%np))
    mr%r(1:mr%np-1) = m%r(1:mr%np-1)
    mr%r(mr%np) = rc

    !Solve system of equations
    call tm_solve_system(mr, rc, rel, qn, ev, rhs, tol, c, ps_g(1:mr%np), ps_f(1:mr%np), ps_v(1:mr%np), .false.)

    !Write info
    write(message(1),'(4x,"Coefficients:")')
    write(message(2),'(6x,"c0  =",1x,es16.9e2,3x,"c2  =",1x,es16.9e2)') c(1), c(2)
    write(message(3),'(6x,"c4  =",1x,es16.9e2,3x,"c6  =",1x,es16.9e2)') c(3), c(4)
    write(message(4),'(6x,"c8  =",1x,es16.9e2,3x,"c10 =",1x,es16.9e2)') c(5), c(6)
    write(message(5),'(6x,"c12 =",1x,es16.9e2)') c(7)
    message(6) = ""
    call write_info(6,20)
    call write_info(6,unit=out_pp)

    !Get the pseudo-wavefunctions and set the quantum numbers to their correct value
    qn%n = qn%l + 1
    ps_g(mr%np:m%np) = ae_g(mr%np:m%np)
    select case (wave_eq)
    case (SCHRODINGER, SCALAR_REL)
      ps_f(mr%np:m%np) = ae_gp(mr%np:m%np)
    case (DIRAC)
      if (rel) then
        ps_f(mr%np:m%np) = ae_f(mr%np:m%np)
      else
        ps_f = M_ZERO
      end if
    end select
    do i = mr%np, m%np
      ps_v(i) = v(ae_potential, m%r(i), qn)
    end do

    !
    deallocate(ae_g, ae_f, ae_gp)
    call mesh_end(mr)

    call pop_sub()
  end subroutine tm_gen

  subroutine tm_equations_rhs(m, rc, rel, qn, ev, grc, gprc, norm, ae_potential, rhs)
    !-----------------------------------------------------------------------!
    ! Returns the right-hand side of the TM set of non-linear equations.    !
    !                                                                       !
    !  m            - mesh                                                  !
    !  rc           - cutoff radius                                         !
    !  rel          - if true, use the relativistic extension to the scheme !
    !  qn           - quantum numbers of valence state                      !
    !  ev           - eigenvalue of valence state                           !
    !  grc          - wave-function at rc                                   !
    !  gprc         - first derivative of the wave-function at rc           !
    !  norm         - norm of the wave-functions up to rc                   !
    !  ae_potential - all-electron potential                                !
    !  rhs          - right-hand side of equations                          !
    !-----------------------------------------------------------------------!
    type(mesh_type),      intent(in)  :: m
    real(R8),             intent(in)  :: rc, ev, grc, gprc, norm
    logical,              intent(in)  :: rel
    type(qn_type),        intent(in)  :: qn
    type(potential_type), intent(in)  :: ae_potential
    real(R8),             intent(out) :: rhs(7)

    integer :: i
    real(R8) :: vrc, vprc, vpprc, vppprc, delta, deltap, deltapp, deltappp, k, alpha, alphap, alphapp
    real(R8), allocatable :: tmp(:)
    type(spline_type) :: spl

    call push_sub("tm_equations_rhs")

    call spline_null(spl)

    !Potential at rc
    vrc   = v(ae_potential, rc, qn)
    vprc  = dvdr(ae_potential, rc, qn)
    vpprc = d2vdr2(ae_potential, rc, qn)

    !Get right hand side of equations to solve
    if (rel) then
      allocate(tmp(m%np))
      do i = 1, m%np
        tmp(i) = dvdr(ae_potential, m%r(i), qn)
      end do
      call spline_init(spl, m%np, m%r, tmp, 3)
      vppprc = spline_eval_deriv(spl, rc)
      call spline_end(spl)
      deallocate(tmp)
      delta    = (vrc - ev)/M_TWO/M_C2
      deltap   = vprc/M_TWO/M_C2
      deltapp  = vpprc/M_TWO/M_C2
      deltappp = vppprc/M_TWO/M_C2

      rhs(1)  = log(grc/rc**(qn%l+1))
      rhs(2)  = gprc/grc - (real(qn%l,r8) + M_ONE)/rc
      k       = -M_TWO*(qn%j - real(qn%l,R8))*(qn%j + M_HALF)
      alpha   = rhs(2) + (real(qn%l,R8) + M_ONE + k)/rc
      rhs(3)  =   M_TWO*(M_ONE - delta)*(vrc - ev) - rhs(2)**2 &
                - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(2) &
                - deltap*alpha/(M_ONE - delta)
      alphap  = rhs(3) - (real(qn%l,R8) + M_ONE + k)/rc**2
      rhs(4)  =   M_TWO*(M_ONE - delta)*vprc &
                - M_TWO*deltap*(vrc - ev) &
                - M_TWO*rhs(2)*rhs(3) &
                + M_TWO*(real(qn%l,r8) + M_ONE)/rc**2*rhs(2) &
                - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(3) &
                - deltapp*alpha/(M_ONE - delta) &
                - deltap**2*alpha/(M_ONE - delta)**2 &
                - deltap*alphap/(M_ONE - delta)
      alphapp = rhs(4) + M_TWO*(real(qn%l,R8) + M_ONE + k)/rc**3
      rhs(5)  =   M_TWO*(M_ONE - delta)*vpprc &
                - M_FOUR*deltap*vprc &
                - M_TWO*deltapp*(vrc - ev) &
                - M_TWO*rhs(2)*rhs(4) &
                - M_TWO*rhs(3)**2 &
                - M_FOUR*(real(qn%l,r8) + M_ONE)/rc**3*rhs(2) &
                + M_FOUR*(real(qn%l,r8) + M_ONE)/rc**2*rhs(3) &
                - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(4) &
                - deltappp*alpha/(M_ONE - delta) &
                - M_THREE*deltap*deltapp*alpha/(M_ONE - delta)**2 &
                - M_TWO*deltap**3*alpha/(M_ONE - delta)**3 &
                - M_TWO*deltapp*alphap/(M_ONE - delta) &
                - M_TWO*deltapp**2*alphap/(M_ONE - delta)**2 &
                - deltap*alphapp/(M_ONE - delta)

    else
      rhs(1) = log(grc/rc**(qn%l+1))
      rhs(2) =   gprc/grc &
               - (real(qn%l,r8) + M_ONE)/rc
      rhs(3) =   M_TWO*(vrc - ev) &
               - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(2) &
               - rhs(2)**2
      rhs(4) =   M_TWO*vprc &
               + M_TWO*(real(qn%l,r8) + M_ONE)/rc**2*rhs(2) &
               - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(3) &
               - M_TWO*rhs(2)*rhs(3)
      rhs(5) =   M_TWO*vpprc &
               - M_FOUR*(real(qn%l,r8) + M_ONE)/rc**3*rhs(2) &
               + M_FOUR*(real(qn%l,r8) + M_ONE)/rc**2*rhs(3) &
               - M_TWO*(real(qn%l,r8) + M_ONE)/rc*rhs(4) &
               - M_TWO*rhs(3)**2 &
               - M_TWO*rhs(2)*rhs(4)
    end if

    rhs(6) = M_ZERO    
    rhs(7) = norm

    call pop_sub()
  end subroutine tm_equations_rhs

  subroutine tm_solve_system(m, rc, rel, qn, ev, rhs, tol, c, ps_g, ps_f, ps_v, silent)
    !-----------------------------------------------------------------------!
    ! Solve the TM set of non-linear equations.                             !
    !                                                                       !
    !  m            - mesh                                                  !
    !  rc           - cutoff radius                                         !
    !  rel          - if true, use the relativistic extension to the scheme !
    !  qn           - quantum numbers of valence state                      !
    !  ev           - eigenvalue of valence state                           !
    !  rhs          - right-hand side of equations                          !
    !  tol          - tolerance                                             !
    !  c            - coeficients of the polynomial                         !
    !  ps_g         - pseudo wave-function                                  !
    !  ps_f         - pseudo wave-function                                  !
    !  ps_v         - pseudo-potential                                      !
    !-----------------------------------------------------------------------!
    type(mesh_type), intent(in)  :: m
    real(R8),        intent(in)  :: rc, ev, rhs(7), tol
    logical,         intent(in)  :: rel, silent
    type(qn_type),   intent(in)  :: qn
    real(R8),        intent(out) :: c(7)
    real(R8),        intent(out) :: ps_g(m%np), ps_f(m%np), ps_v(m%np)

    type(multiroot_solver_type) :: mr_solver
    real(R8)                 :: cc(2)

    call push_sub("tm_solve_system")

    !Guess
    cc = M_ZERO

    !Allocate memory
    call tm_equations_init(m, rc, rel, qn, ev, rhs, silent)

    !Solve non-linear system
    call multiroot_solver_init(2, M_ZERO, M_ZERO, tol, 300, BROYDEN, mr_solver)
    call multiroot_solve_system(mr_solver, cc, tm_equations, tm_equations_write_info)
    c(1) = cc(1)
    c(7) = cc(2)
    call tm_solve_linear_system(c)

    !Compute the final pseudo-wavefunction and the pseudopotential
    call tm_ps_wavefunctions(c, ps_g, ps_f, ps_v)

    !Free memory
    call tm_equations_end()

    call pop_sub()
  end subroutine tm_solve_system

end module troullier_martins
